/*
 * Copyright (c) 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

char copyright[] =
  "@(#) Copyright (c) 1989 Regents of the University of California.\n"
  "All rights reserved.\n";

/*
 * From: @(#)telnetd.c	5.48 (Berkeley) 3/1/91
 */
char telnetd_rcsid[] = 
  "$Id: telnetd.c,v 1.24 2000/04/12 21:36:12 dholland Exp $";

#include "../version.h"

#include <netdb.h>
#include <termcap.h>
#include <netinet/in.h>
/* #include <netinet/ip.h> */ /* Don't think this is used at all here */
#include <arpa/inet.h>
#include <time.h>
#include <assert.h>
#include "telnetd.h"
#include "pathnames.h"
#include "setproctitle.h"

#if	defined(AUTHENTICATE)
#include <libtelnet/auth.h>
#include <libtelnet/auth-proto.h>
#include <libtelnet/misc-proto.h>
int	auth_level = 0;
#endif
#if	defined(SecurID)
int	require_SecurID = 0;
#endif

/* In Linux, this is an enum */
#if defined(__linux__) || defined(IPPROTO_IP)
#define HAS_IPPROTO_IP
#endif

/* secTelnet specific declarations begin */
#define NETBUF_LEN  2048
#define MAX_CERT_BUF_SIZE     2048

#define FIELD_SEPARATOR     ";"

#define TELNET_SVR_VERSION  "1.0"
#define TELNET_SVR_STAMP    "Secure Telnet Server"
#define TELNET_SVR_RAND     "Random Number"
#define TELNET_SVR_CERT      "Switch CSR"

#define MAX_FILENAME_LEN     128
#define COMMAND_LEN	64
#define TELNET_CLI_VERSION "1.0"
#define TELNET_CLI_STAMP "Secure Telnet Client"
#define TELNET_CLI_CERT_REQ  "Requesting Switch CSR"

/* secTelnet specific declarations end */
typedef enum{FALSE, TRUE} Boolean;

static void doit(struct sockaddr_in *who);
static int terminaltypeok(const char *s);
static void setLocalIP(void);
#if 0
static int secValidate(void);
static int secRetrievePEMCert(int sw, char *outbuf, int *outlen);
static int secGetCertificate(int sw, u_char *buf);
static int secGetSystemSecureMode(int sw, int isCP);
#endif

/*
 * I/O data buffers,
 * pointers, and counters.
 */
char	ptyibuf[BUFSIZ], *ptyip = ptyibuf;
char	ptyibuf2[BUFSIZ];

int	hostinfo = 1;			/* do we print login banner? */

int debug = 0;
int keepalive = 1;
char *loginprg = _PATH_LOGIN;
char *progname;

/* Should peer name be resolved when logging in? */
static int	resolve_flag = TRUE;

extern void usage(void);


int
main(int argc, char *argv[], char *env[])
{
	struct sockaddr_in from;
	int on = 1;
	socklen_t fromlen;
	register int ch;

#if	defined(HAS_IPPROTO_IP) && defined(IP_TOS)
	int tos = -1;
#endif

	initsetproctitle(argc, argv, env);

	pfrontp = pbackp = ptyobuf;
	netip = netibuf;
	nfrontp = nbackp = netobuf;
#if	defined(ENCRYPT)
	nclearto = 0;
#endif

	progname = *argv;


	while ((ch = getopt(argc, argv, "d:a:e:lhnr:I:D:B:NsS:a:X:L:")) != EOF) {
		switch(ch) {

#ifdef	AUTHENTICATE
		case 'a':
			/*
			 * Check for required authentication level
			 */
			if (strcmp(optarg, "debug") == 0) {
				extern int auth_debug_mode;
				auth_debug_mode = 1;
			} else if (strcasecmp(optarg, "none") == 0) {
				auth_level = 0;
			} else if (strcasecmp(optarg, "other") == 0) {
				auth_level = AUTH_OTHER;
			} else if (strcasecmp(optarg, "user") == 0) {
				auth_level = AUTH_USER;
			} else if (strcasecmp(optarg, "valid") == 0) {
				auth_level = AUTH_VALID;
			} else if (strcasecmp(optarg, "off") == 0) {
				/*
				 * This hack turns off authentication
				 */
				auth_level = -1;
			} else {
				fprintf(stderr,
			    "telnetd: unknown authorization level for -a\n");
			}
			break;
#endif	/* AUTHENTICATE */

#ifdef BFTPDAEMON
		case 'B':
			bftpd++;
			break;
#endif /* BFTPDAEMON */

		case 'd':
			if (strcmp(optarg, "ebug") == 0) {
				debug++;
				break;
			}
			usage();
			/* NOTREACHED */
			break;

#ifdef DIAGNOSTICS
		case 'D':
			/*
			 * Check for desired diagnostics capabilities.
			 */
			if (!strcmp(optarg, "report")) {
				diagnostic |= TD_REPORT|TD_OPTIONS;
			} else if (!strcmp(optarg, "exercise")) {
				diagnostic |= TD_EXERCISE;
			} else if (!strcmp(optarg, "netdata")) {
				diagnostic |= TD_NETDATA;
			} else if (!strcmp(optarg, "ptydata")) {
				diagnostic |= TD_PTYDATA;
			} else if (!strcmp(optarg, "options")) {
				diagnostic |= TD_OPTIONS;
			} else {
				usage();
				/* NOT REACHED */
			}
			break;
#endif /* DIAGNOSTICS */

#ifdef	AUTHENTICATE
		case 'e':
			if (strcmp(optarg, "debug") == 0) {
				extern int auth_debug_mode;
				auth_debug_mode = 1;
				break;
			}
			usage();
			/* NOTREACHED */
			break;
#endif	/* AUTHENTICATE */

		case 'h':
			hostinfo = 0;
			break;

#ifdef	LINEMODE
		case 'l':
			alwayslinemode = 1;
			break;
#endif	/* LINEMODE */

		case 'L':
			loginprg = strdup(optarg);
			/* XXX what if strdup fails? */
			break;

		case 'n':
			keepalive = 0;
			break;

		case 'N':
			/* Disable name resolution during login */
			resolve_flag = FALSE;
			break;

#ifdef	SecurID
		case 's':
			/* SecurID required */
			require_SecurID = 1;
			break;
#endif	/* SecurID */
		case 'S':
#ifdef	HAS_GETTOS
			if ((tos = parsetos(optarg, "tcp")) < 0)
				fprintf(stderr, "%s%s%s\n",
					"telnetd: Bad TOS argument '", optarg,
					"'; will try to use default TOS");
#else
			fprintf(stderr, "%s%s\n", "TOS option unavailable; ",
						"-S flag not supported\n");
#endif
			break;

#ifdef	AUTHENTICATE
		case 'X':
			/*
			 * Check for invalid authentication types
			 */
			auth_disable_name(optarg);
			break;
#endif	/* AUTHENTICATE */

		default:
			fprintf(stderr, "telnetd: %c: unknown option\n", ch);
			/* FALLTHROUGH */
		case '?':
			usage();
			/* NOTREACHED */
		}
	}


	argc -= optind;
	argv += optind;

	if (debug) {
	    int s, ns;
	    socklen_t foo;
	    struct servent *sp;
	    struct sockaddr_in sn;

	    memset(&sn, 0, sizeof(sn));
	    sn.sin_family = AF_INET;

	    if (argc > 1) {
		usage();
		/* NOTREACHED */
	    } else if (argc == 1) {
		    if ((sp = getservbyname(*argv, "tcp"))!=NULL) {
			sn.sin_port = sp->s_port;
		    } 
		    else {
			int pt = atoi(*argv);
			if (pt <= 0) {
			    fprintf(stderr, "telnetd: %s: bad port number\n",
				    *argv);
			    usage();
			    /* NOTREACHED */
			}
			sn.sin_port = htons(pt);
		   }
	    } else {
		sp = getservbyname("telnet", "tcp");
		if (sp == 0) {
		    fprintf(stderr, "telnetd: tcp/telnet: unknown service\n");
		    exit(1);
		}
		sn.sin_port = sp->s_port;
	    }

	    s = socket(AF_INET, SOCK_STREAM, 0);
	    if (s < 0) {
		    perror("telnetd: socket");;
		    exit(1);
	    }
	    (void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	    if (bind(s, (struct sockaddr *)&sn, sizeof(sn)) < 0) {
		perror("bind");
		exit(1);
	    }
	    if (listen(s, 1) < 0) {
		perror("listen");
		exit(1);
	    }
	    foo = sizeof(sn);
	    ns = accept(s, (struct sockaddr *)&sn, &foo);
	    if (ns < 0) {
		perror("accept");
		exit(1);
	    }
	    (void) dup2(ns, 0);
	    (void) close(ns);
	    (void) close(s);
	} else if (argc > 0) {
		usage();
		/* NOT REACHED */
	}

	openlog("telnetd", LOG_PID | LOG_ODELAY, LOG_DAEMON);
	fromlen = sizeof (from);
	if (getpeername(0, (struct sockaddr *)&from, &fromlen) < 0) {
		fprintf(stderr, "%s: ", progname);
		perror("getpeername");
		_exit(1);
	}
	if (keepalive &&
	    setsockopt(0, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on)) < 0) {
		syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
	}

#if	defined(HAS_IPPROTO_IP) && defined(IP_TOS)
	{
# if	defined(HAS_GETTOS)
		struct tosent *tp;
		if (tos < 0 && (tp = gettosbyname("telnet", "tcp")))
			tos = tp->t_tos;
# endif
		if (tos < 0)
			tos = 020;	/* Low Delay bit */
		if (tos
		   && (setsockopt(0, IPPROTO_IP, IP_TOS, &tos, sizeof(tos)) < 0)
		   && (errno != ENOPROTOOPT) )
			syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
	}
#endif	/* defined(HAS_IPPROTO_IP) && defined(IP_TOS) */
	net = 0;
	doit(&from);
	/* NOTREACHED */
	return 0;
}  /* end of main */

void
usage(void)
{
	fprintf(stderr, "Usage: telnetd");
#ifdef	AUTHENTICATE
	fprintf(stderr, " [-a (debug|other|user|valid|off)]\n\t");
#endif
#ifdef BFTPDAEMON
	fprintf(stderr, " [-B]");
#endif
	fprintf(stderr, " [-debug]");
#ifdef DIAGNOSTICS
	fprintf(stderr, " [-D (options|report|exercise|netdata|ptydata)]\n\t");
#endif
#ifdef	AUTHENTICATE
	fprintf(stderr, " [-edebug]");
#endif
	fprintf(stderr, " [-h]");
#ifdef LINEMODE
	fprintf(stderr, " [-l]");
#endif
	fprintf(stderr, " [-L login_program]");
	fprintf(stderr, " [-n]");
#ifdef	SecurID
	fprintf(stderr, " [-s]");
#endif
#ifdef	AUTHENTICATE
	fprintf(stderr, " [-X auth-type]");
#endif
	fprintf(stderr, " [port]\n");
	exit(1);
}

/*
 * getterminaltype
 *
 *	Ask the other end to send along its terminal type and speed.
 * Output is the variable terminaltype filled in.
 */

static void _gettermname(void);

static
int
getterminaltype(char *name)
{
    int retval = -1;
    (void)name;

    settimer(baseline);

#if defined(AUTHENTICATE)
    /*
     * Handle the Authentication option before we do anything else.
     */
    send_do(TELOPT_AUTHENTICATION, 1);
    while (his_will_wont_is_changing(TELOPT_AUTHENTICATION))
	ttloop();
    if (his_state_is_will(TELOPT_AUTHENTICATION)) {
	retval = auth_wait(name);
    }
#endif

#if	defined(ENCRYPT) 
    send_will(TELOPT_ENCRYPT, 1);
#endif 

	setLocalIP();

    send_do(TELOPT_TTYPE, 1); 
    send_do(TELOPT_TSPEED, 1);
    send_do(TELOPT_XDISPLOC, 1);
    send_do(TELOPT_ENVIRON, 1);

    while (
#if	defined(ENCRYPT)
	   his_do_dont_is_changing(TELOPT_ENCRYPT) ||
#endif
	   his_will_wont_is_changing(TELOPT_TTYPE) ||
	   his_will_wont_is_changing(TELOPT_TSPEED) ||
	   his_will_wont_is_changing(TELOPT_XDISPLOC) ||
	   his_will_wont_is_changing(TELOPT_ENVIRON)) {
	    ttloop();
    }
#if	defined(ENCRYPT)
    /*
     * Wait for the negotiation of what type of encryption we can
     * send with.  If autoencrypt is not set, this will just return.
     */
    if (his_state_is_will(TELOPT_ENCRYPT)) {
	encrypt_wait();
    }
#endif
    if (his_state_is_will(TELOPT_TSPEED)) {
	netoprintf("%c%c%c%c%c%c", 
		   IAC, SB, TELOPT_TSPEED, TELQUAL_SEND, IAC, SE);
    }
    if (his_state_is_will(TELOPT_XDISPLOC)) {
	netoprintf("%c%c%c%c%c%c", 
		   IAC, SB, TELOPT_XDISPLOC, TELQUAL_SEND, IAC, SE);
    }
    if (his_state_is_will(TELOPT_ENVIRON)) {
	netoprintf("%c%c%c%c%c%c", 
		   IAC, SB, TELOPT_ENVIRON, TELQUAL_SEND, IAC, SE);
    }
    if (his_state_is_will(TELOPT_TTYPE)) {
       netoprintf("%c%c%c%c%c%c", 
		  IAC, SB, TELOPT_TTYPE, TELQUAL_SEND, IAC, SE);
    }
    if (his_state_is_will(TELOPT_TSPEED)) {
	while (sequenceIs(tspeedsubopt, baseline))
	    ttloop();
    }
    if (his_state_is_will(TELOPT_XDISPLOC)) {
	while (sequenceIs(xdisplocsubopt, baseline))
	    ttloop();
    }
    if (his_state_is_will(TELOPT_ENVIRON)) {
	while (sequenceIs(environsubopt, baseline))
	    ttloop();
    }
    if (his_state_is_will(TELOPT_TTYPE)) {
	char first[256], last[256];

	while (sequenceIs(ttypesubopt, baseline))
	    ttloop();

	/*
	 * If the other side has already disabled the option, then
	 * we have to just go with what we (might) have already gotten.
	 */
	if (his_state_is_will(TELOPT_TTYPE) && !terminaltypeok(terminaltype)) {
	    /*
	     * Due to state.c, terminaltype points to a static char[41].
	     * Therefore, this assert cannot fail, and therefore, strings
	     * arising from "terminaltype" can be safely strcpy'd into
	     * first[] or last[].
	     */
	    assert(strlen(terminaltype) < sizeof(first));

	    strcpy(first, terminaltype);

	    for(;;) {
		/*
		 * Save the unknown name, and request the next name.
		 */
		strcpy(last, terminaltype);

		_gettermname();
		assert(strlen(terminaltype) < sizeof(first));

		if (terminaltypeok(terminaltype))
		    break;

		if (!strcmp(last, terminaltype) ||
		    his_state_is_wont(TELOPT_TTYPE)) {
		    /*
		     * We've hit the end.  If this is the same as
		     * the first name, just go with it.
		     */
		    if (!strcmp(first, terminaltype))
			break;
		    /*
		     * Get the terminal name one more time, so that
		     * RFC1091 compliant telnets will cycle back to
		     * the start of the list.
		     */
		     _gettermname();
		    assert(strlen(terminaltype) < sizeof(first));

		    if (strcmp(first, terminaltype)) {
			/*
			 * first[] came from terminaltype, so it must fit
			 * back in.
			 */
			strcpy(terminaltype, first);
		    }
		    break;
		}
	    }
	}
    }
    return(retval);
}  /* end of getterminaltype */

static
void
_gettermname(void)
{
    /*
     * If the client turned off the option,
     * we can't send another request, so we
     * just return.
     */
    if (his_state_is_wont(TELOPT_TTYPE))
	return;

    settimer(baseline);
    netoprintf("%c%c%c%c%c%c", IAC, SB, TELOPT_TTYPE, TELQUAL_SEND, IAC, SE);
    while (sequenceIs(ttypesubopt, baseline))
	ttloop();
}

static int
terminaltypeok(const char *s)
{
    /* char buf[2048]; */

    if (terminaltype == NULL)
	return(1);

    /*
     * Fix from Chris Evans: if it has a / in it, termcap will
     * treat it as a filename. Oops.
     */
    if (strchr(s, '/')) {
	return 0;
    }

    /*
     * If it's absurdly long, accept it without asking termcap.
     *
     * This means that it won't get seen again until after login,
     * at which point exploiting buffer problems in termcap doesn't
     * gain one anything.
     *
     * It's possible this limit ought to be raised to 128, but nothing
     * in my termcap is more than 64, 64 is _plenty_ for most, and while
     * buffers aren't likely to be smaller than 64, they might be 80 and
     * thus less than 128.
     */
    if (strlen(s) > 63) {
       return 0;
    }

    /*
     * tgetent() will return 1 if the type is known, and
     * 0 if it is not known.  If it returns -1, it couldn't
     * open the database.  But if we can't open the database,
     * it won't help to say we failed, because we won't be
     * able to verify anything else.  So, we treat -1 like 1.
     */

    /*
     * Don't do this - tgetent is not really trustworthy. Assume
     * the terminal type is one we know; terminal types are pretty
     * standard now. And if it isn't, it's unlikely we're going to
     * know anything else the remote telnet might send as an alias
     * for it.
     *
     * if (tgetent(buf, s) == 0)
     *    return(0);
     */
    return(1);
}

#ifndef	MAXHOSTNAMELEN
#define	MAXHOSTNAMELEN 64
#endif	/* MAXHOSTNAMELEN */

char host_name[MAXHOSTNAMELEN];
char remote_host_name[MAXHOSTNAMELEN];

extern void telnet(int, int);

/*
 * Get a pty, scan input lines.
 */
static void
doit(struct sockaddr_in *who)
{
	const char *host;
	struct hostent *hp = NULL;
	int level;
	char user_name[256];

	/*
	 * Find an available pty to use.
	 */

	pty = getpty();
	if (pty < 0)
		fatal(net, "All network ports in use");

	if (resolve_flag) {
		/* get name of connected client */
		hp = gethostbyaddr((char *)&who->sin_addr,
			sizeof (struct in_addr), who->sin_family);
	}

	if (hp)
		host = hp->h_name;
	else
		host = inet_ntoa(who->sin_addr);

	/*
	 * We must make a copy because Kerberos is probably going
	 * to also do a gethost* and overwrite the static data...
	 */
	{
		int i;
		strncpy(remote_host_name, host, sizeof(remote_host_name)-1);
		remote_host_name[sizeof(remote_host_name)-1] = 0;

		/* Disallow funnies. */
		for (i=0; remote_host_name[i]; i++) {
		    if (remote_host_name[i]<=32 || remote_host_name[i]>126) 
			remote_host_name[i] = '?';
		}
	}
	host = remote_host_name;

	/* Get local host name */
	{
		struct hostent *h = NULL;
		gethostname(host_name, sizeof(host_name));
		if (resolve_flag)
			h = gethostbyname(host_name);
		if (h) {
		    strncpy(host_name, h->h_name, sizeof(host_name));
		    host_name[sizeof(host_name)-1] = 0;
		}
	}

#if	defined(AUTHENTICATE) || defined(ENCRYPT)
	auth_encrypt_init(host_name, host, "TELNETD", 1);
#endif

	init_env();


	/*
	 * get terminal type.
	 */
	*user_name = 0;

 	level = getterminaltype(user_name); 
	setenv("TERM", terminaltype ? terminaltype : "network", 1);
	DIAG(TD_REPORT, netoprintf("td: SECTELNET=%s\n", getenv("SECTELNET")););

	/* TODO list stuff provided by Laszlo Vecsey <master@internexus.net> */

	/*
	 * Set REMOTEHOST environment variable
	 */
	setproctitle("%s", host);
	setenv("REMOTEHOST", host, 0);

	/*
	 * Start up the login process on the slave side of the terminal
	 */
	startslave(host, level, user_name);

	telnet(net, pty);  /* begin server processing */

	/*NOTREACHED*/
}  /* end of doit */

/*
 * Main loop.  Select from pty and network, and
 * hand data to telnet receiver finite state machine.
 */
void telnet(int f, int p)
{
    int on = 1;
    char *HE;
    const char *IM;

    /*
     * Initialize the slc mapping table.
     */
    get_slc_defaults();

    /*
     * Do some tests where it is desireable to wait for a response.
     * Rather than doing them slowly, one at a time, do them all
     * at once.
     */
    if (my_state_is_wont(TELOPT_SGA))
	send_will(TELOPT_SGA, 1);
    /*
     * Is the client side a 4.2 (NOT 4.3) system?  We need to know this
     * because 4.2 clients are unable to deal with TCP urgent data.
     *
     * To find out, we send out a "DO ECHO".  If the remote system
     * answers "WILL ECHO" it is probably a 4.2 client, and we note
     * that fact ("WILL ECHO" ==> that the client will echo what
     * WE, the server, sends it; it does NOT mean that the client will
     * echo the terminal input).
     */
    send_do(TELOPT_ECHO, 1);
    
#ifdef	LINEMODE
    if (his_state_is_wont(TELOPT_LINEMODE)) {
	/*
	 * Query the peer for linemode support by trying to negotiate
	 * the linemode option.
	 */
	linemode = 0;
	editmode = 0;
	send_do(TELOPT_LINEMODE, 1);  /* send do linemode */
    }
#endif	/* LINEMODE */

    /*
     * Send along a couple of other options that we wish to negotiate.
     */
    send_do(TELOPT_NAWS, 1);
    send_will(TELOPT_STATUS, 1);
    flowmode = 1;  /* default flow control state */
    send_do(TELOPT_LFLOW, 1);
    
    /*
     * Spin, waiting for a response from the DO ECHO.  However,
     * some REALLY DUMB telnets out there might not respond
     * to the DO ECHO.  So, we spin looking for NAWS, (most dumb
     * telnets so far seem to respond with WONT for a DO that
     * they don't understand...) because by the time we get the
     * response, it will already have processed the DO ECHO.
     * Kludge upon kludge.
     */
    while (his_will_wont_is_changing(TELOPT_NAWS)) {
	ttloop();
    }
    
    /*
     * But...
     * The client might have sent a WILL NAWS as part of its
     * startup code; if so, we'll be here before we get the
     * response to the DO ECHO.  We'll make the assumption
     * that any implementation that understands about NAWS
     * is a modern enough implementation that it will respond
     * to our DO ECHO request; hence we'll do another spin
     * waiting for the ECHO option to settle down, which is
     * what we wanted to do in the first place...
     */
    if (his_want_state_is_will(TELOPT_ECHO) &&
	his_state_is_will(TELOPT_NAWS)) {
	while (his_will_wont_is_changing(TELOPT_ECHO))
	    ttloop();
    }
    /*
     * On the off chance that the telnet client is broken and does not
     * respond to the DO ECHO we sent, (after all, we did send the
     * DO NAWS negotiation after the DO ECHO, and we won't get here
     * until a response to the DO NAWS comes back) simulate the
     * receipt of a will echo.  This will also send a WONT ECHO
     * to the client, since we assume that the client failed to
     * respond because it believes that it is already in DO ECHO
     * mode, which we do not want.
     */
    if (his_want_state_is_will(TELOPT_ECHO)) {
	DIAG(TD_OPTIONS, netoprintf("td: simulating recv\r\n"););
	willoption(TELOPT_ECHO);
    }
    
    /*
     * Finally, to clean things up, we turn on our echo.  This
     * will break stupid 4.2 telnets out of local terminal echo.
     */
    
    if (my_state_is_wont(TELOPT_ECHO))
	send_will(TELOPT_ECHO, 1);
    
    /*
     * Turn on packet mode
     */
    ioctl(p, TIOCPKT, (char *)&on);
#if defined(LINEMODE) && defined(KLUDGELINEMODE)
    /*
     * Continuing line mode support.  If client does not support
     * real linemode, attempt to negotiate kludge linemode by sending
     * the do timing mark sequence.
     */
    if (lmodetype < REAL_LINEMODE)
	send_do(TELOPT_TM, 1);
#endif	/* defined(LINEMODE) && defined(KLUDGELINEMODE) */
    
    /*
     * Call telrcv() once to pick up anything received during
     * terminal type negotiation, 4.2/4.3 determination, and
     * linemode negotiation.
     */
    telrcv();
    
    ioctl(f, FIONBIO, (char *)&on);
    ioctl(p, FIONBIO, (char *)&on);

#if defined(SO_OOBINLINE)
    setsockopt(net, SOL_SOCKET, SO_OOBINLINE, &on, sizeof on);
#endif	/* defined(SO_OOBINLINE) */
    
#ifdef	SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif
#ifdef	SIGTTOU
    /*
     * Ignoring SIGTTOU keeps the kernel from blocking us
     * in ttioct() in /sys/tty.c.
     */
    signal(SIGTTOU, SIG_IGN);
#endif
    
    signal(SIGCHLD, cleanup);
    
#ifdef TIOCNOTTY
    {
	register int t;
	t = open(_PATH_TTY, O_RDWR);
	if (t >= 0) {
	    (void) ioctl(t, TIOCNOTTY, (char *)0);
	    (void) close(t);
	}
    }
#endif
    
    /*
     * Show banner that getty never gave.
     *
     * We put the banner in the pty input buffer.  This way, it
     * gets carriage return null processing, etc., just like all
     * other pty --> client data.
     */
    
    if (getenv("USER"))
	hostinfo = 0;
    
    IM = DEFAULT_IM;
    HE = 0;

    edithost(HE, host_name);
    if (hostinfo && *IM)
	putf(IM, ptyibuf2);
    
    if (pcc) strncat(ptyibuf2, ptyip, pcc+1);
    ptyip = ptyibuf2;
    pcc = strlen(ptyip);
#ifdef LINEMODE
    /*
     * Last check to make sure all our states are correct.
     */
    init_termbuf();
    localstat();
#endif	/* LINEMODE */

    DIAG(TD_REPORT, netoprintf("td: Entering processing loop\r\n"););
    
    for (;;) {
	fd_set ibits, obits, xbits;
	int c, hifd;
	
	if (ncc < 0 && pcc < 0)
	    break;
	
	FD_ZERO(&ibits);
	FD_ZERO(&obits);
	FD_ZERO(&xbits);
	hifd=0;
	/*
	 * Never look for input if there's still
	 * stuff in the corresponding output buffer
	 */
	if (nfrontp - nbackp || pcc > 0) {
	    FD_SET(f, &obits);
	    if (f >= hifd) hifd = f+1;
	} 
	else {
	    FD_SET(p, &ibits);
	    if (p >= hifd) hifd = p+1;
	}
	if (pfrontp - pbackp || ncc > 0) {
	    FD_SET(p, &obits);
	    if (p >= hifd) hifd = p+1;
	} 
	else {
	    FD_SET(f, &ibits);
	    if (f >= hifd) hifd = f+1;
	}
	if (!SYNCHing) {
	    FD_SET(f, &xbits);
	    if (f >= hifd) hifd = f+1;
	}
	if ((c = select(hifd, &ibits, &obits, &xbits,
			(struct timeval *)0)) < 1) {
	    if (c == -1) {
		if (errno == EINTR) {
		    continue;
		}
	    }
	    sleep(5);
	    continue;
	}
	
	/*
	 * Any urgent data?
	 */
	if (FD_ISSET(net, &xbits)) {
	    SYNCHing = 1;
	}
	
	/*
	 * Something to read from the network...
	 */
	if (FD_ISSET(net, &ibits)) {
#if !defined(SO_OOBINLINE)
	    /*
	     * In 4.2 (and 4.3 beta) systems, the
	     * OOB indication and data handling in the kernel
	     * is such that if two separate TCP Urgent requests
	     * come in, one byte of TCP data will be overlaid.
	     * This is fatal for Telnet, but we try to live
	     * with it.
	     *
	     * In addition, in 4.2 (and...), a special protocol
	     * is needed to pick up the TCP Urgent data in
	     * the correct sequence.
	     *
	     * What we do is:  if we think we are in urgent
	     * mode, we look to see if we are "at the mark".
	     * If we are, we do an OOB receive.  If we run
	     * this twice, we will do the OOB receive twice,
	     * but the second will fail, since the second
	     * time we were "at the mark", but there wasn't
	     * any data there (the kernel doesn't reset
	     * "at the mark" until we do a normal read).
	     * Once we've read the OOB data, we go ahead
	     * and do normal reads.
	     *
	     * There is also another problem, which is that
	     * since the OOB byte we read doesn't put us
	     * out of OOB state, and since that byte is most
	     * likely the TELNET DM (data mark), we would
	     * stay in the TELNET SYNCH (SYNCHing) state.
	     * So, clocks to the rescue.  If we've "just"
	     * received a DM, then we test for the
	     * presence of OOB data when the receive OOB
	     * fails (and AFTER we did the normal mode read
	     * to clear "at the mark").
	     */
	    if (SYNCHing) {
		int atmark;
		
		ioctl(net, SIOCATMARK, (char *)&atmark);
		if (atmark) {
		    ncc = recv(net, netibuf, sizeof (netibuf), MSG_OOB);
		    if ((ncc == -1) && (errno == EINVAL)) {
			ncc = read(net, netibuf, sizeof (netibuf));
			if (sequenceIs(didnetreceive, gotDM)) {
			    SYNCHing = stilloob(net);
			}
		    }
		} 
		else {
		    ncc = read(net, netibuf, sizeof (netibuf));
		}
	    } 
	    else {
		ncc = read(net, netibuf, sizeof (netibuf));
	    }
	    settimer(didnetreceive);
#else	/* !defined(SO_OOBINLINE)) */
	    ncc = read(net, netibuf, sizeof (netibuf));
#endif	/* !defined(SO_OOBINLINE)) */
	    if (ncc < 0 && errno == EWOULDBLOCK)
		ncc = 0;
	    else {
		if (ncc <= 0) {
		    break;
		}
		netip = netibuf;
	    }
	    DIAG((TD_REPORT | TD_NETDATA),
		 netoprintf("td: netread %d chars\r\n", ncc););
	    DIAG(TD_NETDATA, printdata("nd", netip, ncc));
	}
	
	/*
	 * Something to read from the pty...
	 */
	if (FD_ISSET(p, &ibits)) {
	    pcc = read(p, ptyibuf, BUFSIZ);

	    /** BEGIN TEST FOR ENCRYPTION */
#ifdef	TRACE
	    for (i=0; i<pcc; i++) {
		    printf("%x(%c) ", ptyibuf[i], ptyibuf[i]);
		    fflush(stdout);
	    }
	    printf("\n");
#endif

	    /*
	     * Password reading will always flush the input
	     * buffer. We detect it when the terminal device
	     * echo is turned off.
	     */
 	    if (ptyibuf[0] & TIOCPKT_FLUSHREAD) {
 		    struct termios termbuf;

 		    if (tcgetattr(p, &termbuf) != 0) {
			    DIAG(TD_REPORT,
				 netoprintf("td: tcgetattr failed: %s\n",
					    strerror(errno)););
			    break;
		    }
#ifdef	MY_DEBUG
 		    printf("echo=%d, isig=%d, canon=%d\n",
 			   termbuf.c_lflag & ECHO,
 			   termbuf.c_lflag & ISIG,
 			   termbuf.c_lflag & ICANON);
 		    fflush(stdout);
#endif
 		    if (!(termbuf.c_lflag & ECHO) &&
			(termbuf.c_lflag & ICANON) &&
			my_state_is_will(TELOPT_BRCD_ENC)) {
			    DIAG(TD_REPORT,
				 netoprintf("td: start encryption\r\n"););
			    send_do(TELOPT_BRCD_ENC, 1);
 		    }
 	    }

#ifdef	MY_DEBUG
	    if (my_state_is_do(TELOPT_BRCD_ENC)) {
		    printf("DATA=%d, FLUSHREAD=%d, FLUSHWRITE=%d, "
			   "STOP=%d, START=%d, NOSTOP=%d, DOSTOP=%d\n",
			   ptyibuf[0] & TIOCPKT_DATA,
			   ptyibuf[0] & TIOCPKT_FLUSHREAD,
			   ptyibuf[0] & TIOCPKT_FLUSHWRITE,
			   ptyibuf[0] & TIOCPKT_STOP,
			   ptyibuf[0] & TIOCPKT_START,
			   ptyibuf[0] & TIOCPKT_NOSTOP,
			   ptyibuf[0] & TIOCPKT_DOSTOP);
		    fflush(stdout);
	    }
#endif
	    /*
	     * On some systems, if we try to read something
	     * off the master side before the slave side is
	     * opened, we get EIO.
	     */
	    if (pcc < 0 && (errno == EWOULDBLOCK || errno == EIO)) {
		pcc = 0;
	    } 
	    else {
		if (pcc <= 0)
		    break;
#ifdef	LINEMODE
				/*
				 * If ioctl from pty, pass it through net
				 */
		if (ptyibuf[0] & TIOCPKT_IOCTL) {
		    copy_termbuf(ptyibuf+1, pcc-1);
		    localstat();
		    pcc = 1;
		}
#endif	/* LINEMODE */
		if (ptyibuf[0] & TIOCPKT_FLUSHWRITE) {
		    netclear();	/* clear buffer back */
#ifndef	NO_URGENT
		    /*
		     * There are client telnets on some
		     * operating systems get screwed up
		     * royally if we send them urgent
		     * mode data.
		     */
		    netoprintf("%c%c", IAC, DM);
		    neturg = nfrontp-1; /* off by one XXX */
#endif
		}
		if (his_state_is_will(TELOPT_LFLOW) &&
		    (ptyibuf[0] &
		     (TIOCPKT_NOSTOP|TIOCPKT_DOSTOP))) {
			netoprintf("%c%c%c%c%c%c",
				   IAC, SB, TELOPT_LFLOW,
				   ptyibuf[0] & TIOCPKT_DOSTOP ? 1 : 0,
				   IAC, SE);
		}
		pcc--;
		ptyip = ptyibuf+1;
	    }
	}
	
	while (pcc > 0) {
	    if ((&netobuf[BUFSIZ] - nfrontp) < 2)
		break;
	    c = *ptyip++ & 0377, pcc--;
	    if (c == IAC)
		*nfrontp++ = c;
	    *nfrontp++ = c;
	    if ((c == '\r'  ) && (my_state_is_wont(TELOPT_BINARY))) {
		if (pcc > 0 && ((*ptyip & 0377) == '\n')) {
		    *nfrontp++ = *ptyip++ & 0377;
		    pcc--;
		} 
		else *nfrontp++ = '\0';
	    }
	}

	if (FD_ISSET(f, &obits) && (nfrontp - nbackp) > 0)
	    netflush();
	if (ncc > 0)
	    telrcv();
	if (FD_ISSET(p, &obits) && (pfrontp - pbackp) > 0)
	    ptyflush();
    }
    /* In NOS 3.0.1 GA build, login<defunct> process was observed with 
       SQA automated scripts. Following line is added to avoid race 
       condition for wtmp lock */
    signal(SIGCHLD, SIG_IGN);
    cleanup(0);
}  /* end of telnet */
	
#ifndef	TCSIG
# ifdef	TIOCSIG
#  define TCSIG TIOCSIG
# endif
#endif

/*
 * Send interrupt to process on other side of pty.
 * If it is in raw mode, just write NULL;
 * otherwise, write intr char.
 */
void interrupt(void) {
    ptyflush();	/* half-hearted */
    
#ifdef	TCSIG
    (void) ioctl(pty, TCSIG, (char *)SIGINT);
#else	/* TCSIG */
    init_termbuf();
    *pfrontp++ = slctab[SLC_IP].sptr ?
	 (unsigned char)*slctab[SLC_IP].sptr : '\177';
#endif	/* TCSIG */
}

/*
 * Send quit to process on other side of pty.
 * If it is in raw mode, just write NULL;
 * otherwise, write quit char.
 */
void sendbrk(void) {
    ptyflush();	/* half-hearted */
#ifdef	TCSIG
    (void) ioctl(pty, TCSIG, (char *)SIGQUIT);
#else	/* TCSIG */
    init_termbuf();
    *pfrontp++ = slctab[SLC_ABORT].sptr ?
	 (unsigned char)*slctab[SLC_ABORT].sptr : '\034';
#endif	/* TCSIG */
}

void sendsusp(void) {
#ifdef	SIGTSTP
    ptyflush();	/* half-hearted */
# ifdef	TCSIG
    (void) ioctl(pty, TCSIG, (char *)SIGTSTP);
# else	/* TCSIG */
    *pfrontp++ = slctab[SLC_SUSP].sptr ?
	(unsigned char)*slctab[SLC_SUSP].sptr : '\032';
# endif	/* TCSIG */
#endif	/* SIGTSTP */
}

/*
 * When we get an AYT, if ^T is enabled, use that.  Otherwise,
 * just send back "[Yes]".
 */
void recv_ayt(void) {
#if	defined(SIGINFO) && defined(TCSIG)
    if (slctab[SLC_AYT].sptr && *slctab[SLC_AYT].sptr != _POSIX_VDISABLE) {
	(void) ioctl(pty, TCSIG, (char *)SIGINFO);
	return;
    }
#endif
    netoprintf("\r\n[%s : yes]\r\n", host_name);
}

void doeof(void) {
    init_termbuf();

#if	defined(LINEMODE) && (VEOF == VMIN)
    if (!tty_isediting()) {
	extern char oldeofc;
	*pfrontp++ = oldeofc;
	return;
    }
#endif
    *pfrontp++ = slctab[SLC_EOF].sptr ?
	     (unsigned char)*slctab[SLC_EOF].sptr : '\004';
}

/*
 * Function: setLocalIP
 *
 * Descrition:
 *	 This function gets the local server IP address and sets it into
 *	 the environment variable. This infomation is used by login.
 */
static void
setLocalIP(void)
{
	struct sockaddr_in	my;
	socklen_t	len;
	const char	*host;

	DIAG(TD_REPORT, netoprintf("secValidate: calling getsockname\n"););

	len = sizeof (my);
	if (getsockname(net, (struct sockaddr *)&my, &len) < 0) {
		DIAG(TD_REPORT, netoprintf("secValidate: getsockname failed\n"););
		return;
	}
	host = inet_ntoa(my.sin_addr);
	setenv("TELNET_IP_ADDR", host, 1);

	return;
}

#if 0
/* this is specific to brocade code, so defined here */
#define DEV_NAME_LEN	4
#define BRCD_OPTION_LEN	3

/* Function: secValidate
 * Description:
 *		This function sends Brocade specific encryption option 
 *		to Telnet client, if this is secTelnet client, it confirms 
 *		and asks for switch certificate. secTelnet client will validate
 *		this certificate before continuing connection.
 *
 * Note: Use of swInst program and login to CP
 *	swInst -i <ip addr>
 *	This program prints one of the following on the standard output.
 *	sw0, sw1, cp0, cp1
 *
 * 	Above information is used as follows.
 *	1. If user logs into sw0, sw0 certificate is sent to client.
 *	2. If user logs into sw1, sw1 certificate is sent to client.
 *	3. If user logs into cp (any one of the cp), sw0 and sw1 both
 *		must be in non-secure mode for telnet to be allowed. 
 *		If, anyone of switch is in secure mode, sectelnet must
 *		be used to log in. If sectelnet must be used, certificate
 *		of sw0 is sent, if certificate of sw0 does not exist,
 *		certificate of sw1 is sent. Environment variable
 *		SECTELNET is set to include the value 0/1. PAM module reads
 *		it to find out which certificate to use.
 */

static int
secValidate(void)
{
	unsigned char	myopt[] = {IAC, WILL, TELOPT_BRCD_ENC};
	char	netbuf[NETBUF_LEN];
	int	buflen, secure = 0, sw, ret;
	struct sockaddr_in	my;
	FILE	*fp;
	char	command[COMMAND_LEN], dev[DEV_NAME_LEN];
	socklen_t	len;
	const char	*host;
	Boolean		isCP = FALSE;
	struct stat	st;

	DIAG(TD_REPORT, netoprintf("secValidate: entering \n"););

	DIAG(TD_REPORT, netoprintf("secValidate: calling getsockname\n"););

	len = sizeof (my);
	if (getsockname(net, (struct sockaddr *)&my, &len) < 0) {
		DIAG(TD_REPORT, netoprintf("secValidate: getsockname failed\n"););
		return (-1);
	}
	host = inet_ntoa(my.sin_addr);
	setenv("TELNET_IP_ADDR", host, 1);

	/* if fabos is off, telnet is allowed, sectelnet fails */
	if (stat("/sbin/chkconfig", &st) == -1) {
		return (0);	/* Fabos does not exist on platform */
	}

	ret = system("/sbin/chkconfig fabos");

	if (WEXITSTATUS(ret) != 0) { /* fabos is off */
		return (0);
	}

	set_my_want_state_will(TELOPT_BRCD_ENC);

	write(net, myopt, sizeof(myopt));

	/*
	 * have to read fixed number of bytes, if client
	 * sends too fast, you will read next message here
	 * and next read will hang, waiting to read the
	 * data which you have already read. By fixing the
	 * read length, you make sure that next read will get
	 * data intended for it.
	 */
	ncc = read(net, netibuf, BRCD_OPTION_LEN);

	if (ncc < BRCD_OPTION_LEN) {
		DIAG(TD_REPORT, netoprintf("secValidate: read client option,"
					   "length, %d\n", ncc););
		return (-1);
	}

	strncpy(netbuf, netibuf, ncc); /* copy to local buffer */

	DIAG(TD_REPORT, netoprintf("secValidate: received version stamp "
				   "from secTelnet\n"););

	/*
	 * figure out if this for switch 0/1 or cp 0/1
	 */
	snprintf(command, COMMAND_LEN, "/fabos/etc/swInst -i %s", host);

	DIAG(TD_REPORT, netoprintf("secValidate: command to get switch/cp:%s:\n",
				   command););

	fp = popen(command, "r");
	if (fp == NULL) {
		DIAG(TD_REPORT, netoprintf("secValidate: null pipe "
					   "for command\n"););
		return (-1);
	}
	fscanf(fp, "%s", dev);

	if (strncmp("sw0", dev, 3) == 0) {
		DIAG(TD_REPORT, netoprintf("secValidate: comparing "
					   "for sw0, %s\n", dev););
		sw = 0;
	} else if (strncmp("sw1", dev, 3) == 0) {
		DIAG(TD_REPORT, netoprintf("secValidate: comparing "
					   "for sw1, %s\n", dev););
		sw = 1;
	} else if (strncmp("cp", dev, 2) == 0) {
		DIAG(TD_REPORT, netoprintf("secValidate: comparing "
					   "for cp, %s\n", dev););
		isCP = TRUE;
		sw = 0;
	} else {
		DIAG(TD_REPORT, netoprintf("secValidate: ip could not "
					   "be mapped to device, %s\n", dev););
		return (-1);
	}

	if (pclose(fp) == -1) {
		DIAG(TD_REPORT, netoprintf("secValidate: pclose failed\n"););
		return (-1);
	}

	if ((secure = secGetSystemSecureMode(sw, isCP)) < 0) {
		DIAG(TD_REPORT, netoprintf("secValidate: "
					   "secure mode,%d\n", secure););
		return (-1);
	}

	DIAG(TD_REPORT, netoprintf("secValidate: secure mode,%d\n", secure););

	if (netbuf[1] != DO || netbuf[2] != TELOPT_BRCD_ENC) {
		/* normal telnet */
		if (secure) {
			/* can not log into the secured switch */
			DIAG(TD_REPORT, netoprintf("secValidate: login "
						   "from normal telnet\n"););
			bzero(netbuf, NETBUF_LEN);
			strcpy(netbuf, "Can only use secure telnet ");
			strcat(netbuf, "to log into this security ");
			strcat(netbuf, "enabled switch.\r\n");
			buflen = strlen(netbuf) + 1;
			write(net, netbuf, buflen);
			return (-1);
		} else { /* normal telnet is allowed */
			netip = netibuf;
			set_my_state_wont(TELOPT_BRCD_ENC);
			telrcv();
			if (ncc > 0) {
				pfrontp = pbackp = ptyobuf;
				telrcv();
			}
			return (0);
		}
	}

        /* Now read client version stamp */
	buflen = read(net, netbuf, NETBUF_LEN);
	if (buflen <= 0) {
		return (-1);
	}

	{
		/* Certificate is installed */
		static char	cert[MAX_CERT_BUF_SIZE];
		char	*cliver;
		int	session_random_no, certlen;
		char	str_rand[16];
		char	str_env[24];

		cliver = strstr(netbuf, TELNET_CLI_VERSION);
		if (!cliver || !strstr(netbuf, TELNET_CLI_STAMP)) {
			if (!secure){ /* malformed request */
				return (-1);
			}

		}

		if (strncmp(cliver, TELNET_SVR_VERSION,
			    sizeof (TELNET_SVR_VERSION) - 1)) {
			DIAG(TD_REPORT, netoprintf("secValidate: telnet "
						   "client version mismatch\n"););
			return (-1);
		}

		srand(time(NULL));
		session_random_no = rand();
		snprintf(str_rand, sizeof(str_rand)-1, "%d",
			 session_random_no);
		snprintf(netbuf, sizeof(netbuf)-1,
			 "%s=%d", TELNET_SVR_RAND, session_random_no);
		buflen = strlen (netbuf) + 1;
		if (write(net, netbuf, buflen) < buflen) {
			return (-1);
		}
		buflen = read(net, netbuf, NETBUF_LEN);
		if (buflen <= 0) {
			DIAG(TD_REPORT,
			     netoprintf("secValidate:"
					"cert request read failed\n"););
			return (-1);
		}
		if (!strstr(netbuf, TELNET_CLI_CERT_REQ)) {
			DIAG(TD_REPORT, netoprintf("secValidate:"
						   "invalid cert request\n"););
			return (-1);
		}

		/* get certificate for switch */

		ret = secRetrievePEMCert(sw, cert, &certlen);

		/* login through CP but cert for sw0 not found,
		 * try with certificate of sw1, if this also
		 * fails, return error.
		 */
		if (isCP && ret != 0) {
			DIAG(TD_REPORT,
			     netoprintf("secValidate: Cp login, "
					"sw0 cert missing\n"););
			
			if (sw == 0) sw = 1;
			else sw = 0;
			ret = secRetrievePEMCert(sw, cert, &certlen);
			if (ret < 0) {
				DIAG(TD_REPORT,
				     netoprintf("secValidate: no switch "
						"certs found\n"););
				return (-1);
			}
		} else if (ret < 0) {
			DIAG(TD_REPORT, netoprintf("DEBUG: Sw login, "
						   "no cert found\n"););
			return (-1);
		}

		strcpy(netbuf, TELNET_SVR_CERT);
		strcat(netbuf, "=");
		strcat(netbuf, cert);
		strcat(netbuf, "\n");
		buflen = strlen(netbuf) + 1;
		if (write(net, netbuf, buflen) < buflen) {
			DIAG(TD_REPORT, netoprintf("secValidate:"
						   "certificate write failed\n"););
			return (-1);
		}

		snprintf(str_env, sizeof(str_env), "%s:%d", str_rand, sw);

		DIAG(TD_REPORT, netoprintf("secValidate: SECTELNET %s\n", str_env););

		/* set the environment variable for secTelnet */
		if (setenv("SECTELNET", str_env, 1) != 0) {
			DIAG(TD_REPORT, netoprintf("failed to set env SECTELNET\n"););
			return (-1);
		}

		DIAG(TD_REPORT, netoprintf("secValidate: "
					   "set my state will %d\n",
					   TELOPT_BRCD_ENC););
		set_my_state_will(TELOPT_BRCD_ENC);
	}

	DIAG(TD_REPORT, netoprintf("secValidate: exiting \n"););
	return (0);
}

/*
 * Function: secRetrievePEMCert
 *
 * Description:
 *  converts base64 format certificate retrieved from
 *  flash to PEM format and returns it in caller supplied
 *  buffer. Caller must allocate this buffer.
 *
 * Input:
 * Output:
 *	outbuf  PEM format certificate
 *	outlen  length of certificate
 * Return: (0) - success (-1) - failure
 *
 */

int
secRetrievePEMCert(int sw, char *outbuf, int *outlen)
{
	char	buf[MAX_CERT_BUF_SIZE];
	int	len, i, pem_len;

	if (secGetCertificate(sw, buf) < 0) {
		DIAG(TD_REPORT, netoprintf("td: failed to get certificate\n"));
		return (-1);
	}

	strcpy(outbuf, "-----BEGIN CERTIFICATE-----");

	len = strlen(buf);
	pem_len = strlen(outbuf);

	for (i = 0; i < len; i++) {
		if ((i % 64) == 0) {
			outbuf[pem_len++] = '\n';
		}
		outbuf[pem_len++] = buf[i];
	}
	outbuf[pem_len] = '\0';

	strcat(outbuf, "\n-----END CERTIFICATE-----");
	*outlen = strlen(outbuf);

	return (0);
}

/*
 * Function: secGetCertificate
 *
 * Description:
 *	Get certificate from flash
 * Input:
 * Output:
 *  outbuf  switch certificate 
 * Return: length of object, (-1) - failure
 *
 */

#undef FUNCTION
#define FUNCTION	"secGetCertificate"
#define CERTFILE	"/etc/fabos/pki/switch.%d.crt"

int
secGetCertificate(int sw, u_char *buf)
{
	char	pfile[PATH_MAX];
	FILE	*fp;
	struct stat	st;
	u_int	len;

	DIAG(TD_REPORT, netoprintf("secGetCertifiate: Entering\n"););

	snprintf(pfile, sizeof (pfile), CERTFILE, sw);
	fp = fopen(pfile, "r");
	if (!fp) {
		DIAG(TD_REPORT,
		     netoprintf("secGetCertificate:"
				"Certificate file open failed\n"););
		return (-1);
	}

	if (stat(pfile, &st) == -1) {
		DIAG(TD_REPORT,
		     netoprintf("secGetCertifiate: failed to read "
				"certificate file stat\n"););
		return (-1);
	}
	len = st.st_size;

	if (fread(buf, len, 1, fp) != 1) {
		DIAG(TD_REPORT, netoprintf("secGetCertifiate: exiting, success\n"););
		return (-1);
	}

	buf[len] = '\0';
	fclose(fp);
	DIAG(TD_REPORT, netoprintf("secGetCertifiate: exiting, success\n"););
	return (len);
}

/*
 * Function: secGetSystemSecureMode
 *
 * Description:
 *	Get system wide secure mode,
 *	If any of the switch in chassis is in secure
 *	mode, system is considered to be in secure
 *	mode
 * Input:
 * Output:
 *  
 * Return: 1 - system in secure mode
 *	   0 - system in non-secure mode
 *	   -1 - Error
 *
 */

static int
secGetSystemSecureMode(int sw, int isCP)
{
	int	sw_local, secure;
	FILE	*fp;
	char	command[COMMAND_LEN];

	sw_local = sw;
	snprintf(command, COMMAND_LEN, "/fabos/bin/secModeGet %d", sw_local);
	DIAG(TD_REPORT, netoprintf("secModeget command1, %s\n", command););
	fp = popen(command, "r");
	if (fp == NULL) {
		DIAG(TD_REPORT, netoprintf("failed to open secModeGet1\n"););
		return (-1);
	}
	/* 1 - secure mode, 0 - non-secure mode */
	fscanf(fp, "%d", &secure);
	DIAG(TD_REPORT, netoprintf("sec mode1, %d\n", secure););

/*  	if (secure != 0 && secure != 1) { */
/*  		DIAG(TD_REPORT, netoprintf("invalid sec mode value1\n");); */
/*  		return (-1); */
/*  	} */

	if (pclose(fp) == -1) {
		DIAG(TD_REPORT, netoprintf("secGetSystemSecureMode:"
					   "pclose failed1\n"););
		return (-1);
	}

	if (isCP && !secure) {
		if (sw_local == 0) {
			sw_local = 1;
		} else {
			sw_local = 0;
		}
	} else {
		DIAG(TD_REPORT, netoprintf("returning secmode, %d\n", secure););
		return (secure);
	}

	snprintf(command, COMMAND_LEN, "/fabos/bin/secModeGet %d", sw_local);
	DIAG(TD_REPORT, netoprintf("secModeget command2, %s\n", command););
	fp = popen(command, "r");
	if (fp == NULL) {
		DIAG(TD_REPORT, netoprintf("failed to open secModeGet2\n"););
		return (-1);
	}
	/* 1 - secure mode, 0 - non-secure mode */
	fscanf(fp, "%d", &secure);

	if (secure != 0 && secure != 1) {
		DIAG(TD_REPORT, netoprintf("invalid sec mode value2\n"););
		return (-1);
	}

	if (pclose(fp) == -1) {
		DIAG(TD_REPORT, netoprintf("secGetSystemSecureMode:"
					   "pclose failed2\n"););
		return (-1);
	}

	DIAG(TD_REPORT, netoprintf("secGetSystemSecureMode: exiting\n"););

	return (secure);
}
#endif
